MODULE bdrymod

  ! Loads
  ! altimetry (top bdry) and bathymetry (bottom bdry) data
  !
  ! x = coordinate of boundary
  ! t = tangent for a segment
  ! n = normal  for a segment
  ! Rlen = length of tangent (temporary variable to normalize tangent)
  ! tangents are also calculated for the nodes, if the curvilinear option is used for the boundaries.

  IMPLICIT NONE
  SAVE
  INTEGER, PARAMETER :: ATIFIL = 40, BTYFIL = 41
  INTEGER NATIPTS, NBTYPTS, ii, I, IOStat, IAllocStat
  CHARACTER*1 atiType, btyType
  REAL (KIND=8), ALLOCATABLE :: xBot( :, : ), tBot( :, : ), nBot( :, : ), tBotNode( :, : ), nBotNode( :, : ), &
                                RLenBot( : ), kappaBot( : ), &
                                xTop( :, : ), tTop( :, : ), nTop( :, : ), tTopNode( :, : ), nTopNode( :, : ), &
                                RLenTop( : ), kappaTop( : )

CONTAINS

  SUBROUTINE READATI( TopATI, DepthT, rBox, PRTFil )

    IMPLICIT NONE
    INTEGER     PRTFil
    CHARACTER*1 TopATI
    REAL DepthT, rBox
    REAL (KIND=8) sss
    REAL (KIND=8), ALLOCATABLE :: phi( : )

    IF ( TopATI == '*' ) THEN
       WRITE( PRTFIL, * ) '*********************************'
       WRITE( PRTFIL, * ) 'Using top-altimetry file'

       OPEN( UNIT = ATIFil,   FILE = 'ATIFIL', STATUS = 'OLD', IOSTAT = IOStat )
       IF ( IOsTAT /= 0 ) CALL ERROUT( PRTFIL, 'F', 'ReadATI', 'Unable to open altimetry file' )

       READ(  ATIFIL, * ) atiType
       SELECT CASE ( atiType )
       CASE ( 'C' )
          WRITE( PRTFil, * ) 'Curvilinear Interpolation'
       CASE ( 'L' )
          WRITE( PRTFil, * ) 'Piecewise linear interpolation'
       CASE DEFAULT
          CALL ERROUT( PRTFil, 'F', 'ReadATI', 'Unknown option for selecting altimetry interpolation' )
       END SELECT

       READ(  ATIFIL, * ) NatiPts
       WRITE( PRTFIL, * ) 'Number of altimetry points', NatiPts
       NatiPts = NatiPts + 2   ! we'll be extending the altimetry to infinity to the left and right

       ALLOCATE( xTop(  2, NatiPts ), tTop(  2, NatiPts ), nTop( 2, NatiPts ), &
            tTopNode( 2, NatiPts ), nTopNode( 2, NatiPts ), &
            RLenTop( NatiPts - 1 ), phi( NatiPts ), kappaTop( NatiPts - 1 ), Stat = IAllocStat )
       IF ( IAllocStat /= 0 ) &
            CALL ERROUT( PRTFIL, 'F', 'BELLHOP:READATI', 'Insufficient memory for altimetry data: reduce # ati points' )

       WRITE( PRTFIL, * )
       WRITE( PRTFIL, * ) ' Range (km)  Depth (m)'

       DO I = 2, NatiPts - 1
          READ(  ATIFIL, * ) xTop( :, I )
          WRITE( PRTFIL, FMT = "(2G11.3)" ) xTop( :, I )
          IF ( xTop( 2, I ) < DepthT ) THEN
             CALL ERROUT( PRTFIL, 'F', 'BELLHOP:READATI', 'Altimetry rises above highest point in the sound speed profile' )
          END IF
       END DO

       CLOSE( ATIFIL )

       xTop( 1, : ) = 1000.0 * xTop( 1, : )   ! Convert ranges in km to m

       ! extend the altimetry to +/- infinity in a piecewise constant fashion

       xTop( 1, 1 )       = -sqrt( huge( XTop( 1, 1 ) ) ) / 1.0d5
       xTop( 2, 1 )       = xTop( 2, 2 )
       xTop( 1, NatiPts ) = +sqrt( huge( xTop( 1, 1 ) ) ) / 1.0d5
       xTop( 2, NatiPts ) = xTop( 2, NatiPts - 1 )

       ! compute tangent and outward-pointing normal to top
       !tToP( 1, : ) = xTop( 1, 2:NatiPts ) - xTop( 1, 1:NatiPts - 1 )
       !tTop( 2, : ) = xTop( 2, 2:NatiPts ) - xTop( 2, 1:NatiPts - 1 )

       DO ii = 1, NatiPts - 1
          tTop( :, ii ) = xTop( :, ii+1 ) - xTop( :, ii )
          RLenTop( ii ) = SQRT( tTop( 1, ii ) ** 2 + tTop( 2, ii ) ** 2 )
          tTop( :, ii ) = tTop( :, ii ) / RLenTop( ii )

          nTop( 1, ii ) =  tTop( 2, ii )
          nTop( 2, ii ) = -tTop( 1, ii )
       END DO

       IF ( atiType == 'C' ) THEN ! curvilinear option: compute tangent and normal at node by averaging normals on adjacent segments

          DO ii = 2, NatiPts - 1
 
             sss = RLenTop( ii - 1 ) / ( RLenTop( ii - 1 ) + RLenTop( ii ) )
             sss = 0.5
             tTopNode( :, ii ) = ( 1.0 - sss ) * tTop( :, ii - 1 ) + sss * tTop( :, ii )

             !tTopNode( :, ii ) = ( RLenTop( ii ) * tTop( :, ii - 1 ) + RlenTop( ii - 1 ) * tTop( :, ii ) ) / &
             !                    ( RLenTop( ii )                     + RLenTop( ii - 1 ) )
          END DO

          tTopNode( :, 1       ) = (/ 1.0, 0.0 /)   ! tangent left-end  node
          tTopNode( :, NatiPts ) = (/ 1.0, 0.0 /)   ! tangent right-end node
          nTopNode( 1, : ) = +tTopNode( 2, : )
          nTopNode( 2, : ) = -tTopNode( 1, : )

          ! compute curvature in each segment
          phi = atan2( tTopNode( 2, : ), tTopNode( 1, : ) )
          DO ii = 1, NatiPts - 2
             kappaTop( ii ) = ( phi( ii + 1 ) - phi( ii ) ) / RLenTop( ii ) ! this is curvature = dphi/ds
          END DO 

       ELSE
          kappaTop = 0
       END IF

    ELSE   ! no bathymetry given, use SSP depth for flat top
       NatiPts = 2
       atiType = 'L'
       ALLOCATE( xTop( 2, 2 ), tTop( 2, 2 ), nTop( 2, 2 ), RLenTop( 1 ), kappaTop( 1 ), Stat = IAllocStat )
       IF ( IAllocStat /= 0 ) CALL ERROUT( PRTFIL, 'F', 'BELLHOP', 'Insufficient memory'  )
       xTop( :, 1 ) = (/ -rBox, DepthT /)
       xTop( :, 2 ) = (/  rBox, DepthT /)

       tTop( :, 1 ) = (/ 1.0,  0.0 /)   ! tangent to top
       nTop( :, 1 ) = (/ 0.0, -1.0 /)   ! outward-pointing normal
       tTop( :, 2 ) = (/ 1.0,  0.0 /)   ! tangent to top
       nTop( :, 2 ) = (/ 0.0, -1.0 /)   ! outward-pointing normal

       RLenTop = sqrt( tTop( 1, 1 )** 2 + tTop( 2, 1 )** 2 )

       kappaTop( 1 ) = 0

    ENDIF

  END SUBROUTINE READATI

  ! **********************************************************************!

  SUBROUTINE READBTY( BotBTY, DepthB, rBox, PRTFil )

    ! Reads in the bottom bathymetry

    IMPLICIT NONE
    INTEGER     PRTFil
    CHARACTER*1 BotBTY
    REAL DepthB, rBox
    REAL (KIND=8) sss
    REAL (KIND=8), ALLOCATABLE :: phi( : )

    IF ( BotBTY == '*' ) THEN
       WRITE( PRTFIL, * ) '*********************************'
       WRITE( PRTFIL, * ) 'Using bottom-bathymetry file'

       OPEN( UNIT = BTYFil,   FILE = 'BTYFIL', STATUS = 'OLD', IOSTAT = IOStat )
       IF ( IOsTAT /= 0 ) CALL ERROUT( PRTFIL, 'F', 'ReadBTY', 'Unable to open bathymetry file' )
 
       READ( BTYFIL, * ) btyType
 
       SELECT CASE ( btyType )
       CASE ( 'C' )
          WRITE( PRTFil, * ) 'Curvilinear Interpolation'
       CASE ( 'L' )
          WRITE( PRTFil, * ) 'Piecewise linear interpolation'
       CASE DEFAULT
          CALL ERROUT( PRTFil, 'F', 'ReadBTY', 'Unknown option for selecting bathymetry interpolation' )
       END SELECT

       READ(  BTYFIL, * ) NbtyPts
       WRITE( PRTFIL, * ) 'Number of bathymetry points', NbtyPts

       NbtyPts = NbtyPts + 2   ! we'll be extending the bathymetry to infinity on both sides
       ALLOCATE( xBot( 2, NbtyPts ), tBot( 2, NbtyPts - 1 ), nBot( 2, NbtyPts - 1 ), &
            tBotNode( 2, NbtyPts ), nBotNode( 2, NbtyPts ), &
            RlenBot( NbtyPts - 1 ), phi( NbtyPts ), kappaBot( NbtyPts - 1 ), Stat = IAllocStat )
       IF ( IAllocStat /= 0 ) &
            CALL ERROUT( PRTFIL, 'F', 'BELLHOP:READBTY', 'Insufficient memory for bathymetry data: reduce # bty points' )

       WRITE( PRTFIL, * )
       WRITE( PRTFIL, * ) ' Range (km)  Depth (m)'

       DO I = 2, NbtyPts - 1
          READ(  BTYFIL, * ) xBot( :, I )
          WRITE( PRTFIL, FMT = "(2G11.3)" ) xBot( :, I )
          IF ( xBot( 2, I ) > DepthB ) THEN
             CALL ERROUT( PRTFIL, 'F', 'BELLHOP:READBTY', 'Bathymetry drops below lowest point in the sound speed profile' )
          END IF
 
       END DO

       CLOSE( BTYFIL )

       xBot( 1, : ) = 1000.0 * xBot( 1, : )   ! Convert ranges in km to m

       ! extend the bathymetry to +/- infinity in a piecewise constant fashion

       xBot( 1, 1 )       = -sqrt( huge( xBot( 1, 1 ) ) ) / 1.0d5
       xBot( 2, 1 )       = xBot( 2, 2 )
       xBot( 1, NbtyPts ) = +sqrt( huge( xBot( 1, 1 ) ) ) / 1.0d5
       xBot( 2, NbtyPts ) = xBot( 2, NbtyPts - 1 )

       ! compute tangent and outward-pointing normal to each bottom segment
       !tBot( 1, : ) = xBot( 1, 2:NbtyPts ) - xBot( 1, 1:NbtyPts - 1 )
       !tBot( 2, : ) = xBot( 2, 2:NbtyPts ) - xBot( 2, 1:NbtyPts - 1 )
       ! above caused compiler problems

       DO ii = 1, NbtyPts - 1
          tBot( :, ii ) = xBot( :, ii+1 ) - xBot( :, ii )
          RLenBot( ii ) = SQRT( tBot( 1, ii ) ** 2 + tBot( 2, ii ) ** 2 )
          tBot( :, ii ) =  tBot( :, ii ) / RLenBot( ii )

          nBot( 1, ii ) = -tBot( 2, ii )
          nBot( 2, ii ) = +tBot( 1, ii )
       END DO

       IF ( btyType == 'C' ) THEN ! curvilinear option: compute tangent and normal at node by averaging normals on adjacent segments

          DO ii = 2, NbtyPts - 1
             sss = RLenBot( ii - 1 ) / ( RLenBot( ii - 1 ) + RLenBot( ii ) )
             sss = 0.5
             tBotNode( :, ii ) = ( 1.0 - sss ) * tBot( :, ii - 1 ) + sss * tBot( :, ii )
          END DO

          tBotNode( :, 1       ) = (/ 1.0, 0.0 /)   ! tangent left-end  node
          tBotNode( :, NbtyPts ) = (/ 1.0, 0.0 /)   ! tangent right-end node
          nBotNode( 1, : ) = -tBotNode( 2, : )
          nBotNode( 2, : ) = +tBotNode( 1, : )

          ! compute curvature in each segment
          phi = atan2( tBotNode( 2, : ), tBotNode( 1, : ) )   ! this is the angle at each node
          DO ii = 1, NbtyPts - 2
             kappaBot( ii ) = ( phi( ii + 1 ) - phi( ii ) ) / RLenBot( ii ) ! this is curvature = dphi/ds
          END DO

       ELSE
          kappaBot = 0
       END IF

    ELSE   ! no bathymetry given, use SSP depth for flat bottom
       btyType = 'L'
       NbtyPts = 2
       ALLOCATE( xBot( 2, 2 ), tBot( 2, 2 ), nBot( 2, 2 ), RLenBot( 1 ), kappaBot( 1 ), Stat = IAllocStat )
       IF ( IAllocStat /= 0 ) CALL ERROUT( PRTFIL, 'F', 'BELLHOP', 'Insufficient memory'  )
       xBot( :, 1 ) = (/ -rBox, DEPTHB /)
       xBot( :, 2 ) = (/  rBox, DEPTHB /)

       tBot( :, 1 ) = (/ 1.0, 0.0 /)   ! tangent to bottom
       nBot( :, 1 ) = (/ 0.0, 1.0 /)   ! outward-pointing normal
       tBot( :, 2 ) = (/ 1.0, 0.0 /)   ! tangent to bottom
       nBot( :, 2 ) = (/ 0.0, 1.0 /)   ! outward-pointing normal

       RLenBot = sqrt( tBot( 1, 1 )** 2 + tBot( 2, 1 )** 2 )
       kappaBot( 1 ) = 0

    ENDIF

  END SUBROUTINE READBTY

END MODULE bdrymod
