PROGRAM KRAKEL

  ! Program for solving for ocean acoustic normal modes
  ! Michael B. Porter

  USE krakelmod
  
  IMPLICIT REAL (KIND=8) (A-H, O-Z)

  NV( 1:5 ) = (/ 1, 2, 4, 8, 16 /)
  CALL GETPAR
  M = MaxM

  WRITE( PRTFIL, * )
  WRITE( PRTFIL, * ) 'Mesh multiplier   CPU seconds'

  ! Main loop: solve the problem for a sequence of meshes

  DO ISET = 1, NSETS

     N( 1:NMedia ) = NG( 1:NMedia ) * NV( ISet )
     H( 1:NMedia ) = ( Depth( 2:NMedia + 1 ) - Depth( 1:NMedia ) ) / N( 1:NMedia )
     HV( ISet ) = H( 1 )

     CALL SOLVE( ERROR )
     IF ( ERROR * 1000.0 * RMax < 1.0 ) GOTO 3000
  END DO

  ! Fall through indicates failure to converge
  WRITE( PRTFIL, * ) 'WARNING: Too many meshes needed, check convergence'

  ! Solution complete: discard modes with phase velocity above CHIGH

3000 omega = SQRT(omega2)
  DO MODE = M, 1, -1
     IF ( omega/DBLE( SQRT( EXTRAP( 1, MODE ) ) ) < CHIGH ) GOTO 5000
  END DO
5000 M = MODE

  ! Write eigenvalues to PRTFIL and MODFIL
  WRITE( PRTFIL, * )
  WRITE( PRTFIL, * ) '   I           K                ALPHA         PHASE SPEED'

  k( 1:M ) = SQRT( Extrap( 1, 1:M ) + k( 1:M ) )
  DO Mode = 1, M
     WRITE( PRTFil, "( I5, 3G18.10 )" ) Mode, k( Mode ), omega / DBLE( k( Mode ) )
  END DO

  WRITE( MODFIL, REC=5 ) M, LRECL
  IFIRST = 1
  DO IREC = 1, 1 + ( 2 * M - 1 ) / LRECL

     ILAST  = MIN( M, IFIRST + LRECL / 2 - 1 )
     WRITE( MODFIL, REC = 6 + M + IREC ) CMPLX( k( IFIRST : ILAST ) )                            

     IFIRST = ILAST + 1
  END DO

  STOP
END PROGRAM KRAKEL
!**********************************************************************C
SUBROUTINE GETPAR

  ! Read in the ENVFIL data

  USE krakelmod
  USE SdRdRMod
  USE RefCoMod

  IMPLICIT REAL (KIND=8) (A-H, O-Z)
  REAL (KIND=4) :: ZMin, ZMax

  ! Read in environmental info

  TITLE = 'KRAKEL-  '

  CALL READIN( Title, Freq, MaxMedium, NMedia, &
       TopOpt, cpT, csT, rhoT, BumDen, eta, xi, NG, Sigma, Depth, &
       BotOpt, cpB, csB, rhoB, ENVFil, PRTFil )

  READ(  ENVFil, *    ) CLow, CHigh                 ! Spectral limits
  WRITE( PRTFil, "( /, ' CLow = ', G12.5, '  CHigh = ', G12.5 )" ) CLow, CHigh
  IF ( CLow <= 0.0 .OR. CHigh <= 0.0 .OR. CLow >= CHigh ) &
       CALL ERROUT( PRTFil, 'F', 'GETPAR', 'Need phase speeds CLow, CHigh > 0 and CLow < CHigh'  )

  READ(  ENVFil, * ) RMax                           ! Maximum range for calculations
  WRITE( PRTFil, * ) 'RMax = ', RMax

  ! Read source/receiver depths
  ZMin = Depth( 1 )
  ZMax = Depth( NMedia + 1 )
  CALL SDRD( ENVFil, PRTFil, ZMin, ZMax )

  CLOSE( ENVFIL )
  omega2 = ( 2.0 * PI * Freq ) ** 2

  CALL ReadRC( BotOpt(1:1), TopOpt(2:2), PRTFil )   ! Optionally read in Bot, Top reflection coefficients
  M = MaxM

  RETURN
END SUBROUTINE GETPAR
!**********************************************************************C
SUBROUTINE INIT

  ! Initializes arrays defining difference equations
  ! Identical to KRAKENC apart for B1, B2, B3, B4 and rho

  USE krakelmod

  IMPLICIT REAL (KIND=8) (A-H, O-Z)
 
  LOGICAL :: ELFLAG = .FALSE.
  COMPLEX (KIND=8) :: CP2, CS2
  COMPLEX (KIND=8), ALLOCATABLE :: CP( : ), CS( : )
  CHARACTER Task*8

  REAL    (KIND=8) :: cMinV
  COMPLEX (KIND=8) :: LAM, MU

  CMIN = 1.0E6
  NFACT = 0
  Loc(1) = 0

  ! Allocate storage for finite-difference coefficients

  NPTS = SUM( N( 1:NMedia ) ) + NMedia

  IF ( ALLOCATED( B1 ) ) DEALLOCATE( B1, B1C, B2, B3, B4, rho )
  ALLOCATE ( B1( NPTS ), B1C( NPTS ), B2( NPTS ), B3( NPTS ), B4( NPTS ), rho( NPTS ), &
       CP( NPTS ), CS( NPTS ), Stat = IAllocStat )
  IF ( IAllocStat /= 0 ) &
       CALL ERROUT( PRTFil, 'F', 'KRAKEN - INIT', 'Insufficient memory: Reduce mesh.' )

  J = 0

  DO Medium = 1, NMedia   ! Loop over media
     IF ( Medium /= 1 ) Loc(Medium) = Loc(Medium-1) + N(Medium-1) + 1
     N1 = N( Medium ) + 1

     IF ( Loc( Medium ) + N1 > MaxN ) THEN
        WRITE( PRTFIL, * ) 'Insufficient storage for next mesh'
        WRITE( PRTFIL, * ) 'PROGRAM ABORTING'
        STOP 'ERROR IN KRAKENC: Insufficient storage for next mesh'
     ENDIF

     CALL PROFIL( DEPTH, CP, CS, rho( Loc(Medium)+1 ), Medium, N1, FREQ, TOPOPT(1:1), TOPOPT(3:3), 'TAB' )

     IF ( CS(1) == (0.0, 0.0) ) THEN

        ! Case of an acoustic medium

        Mater(Medium) = 'ACOUSTIC'
        IF ( NFACT == 0 ) NFACT = Medium
        NLACT = Medium

        DO II = 1, N1
           CMIN = MIN( DBLE( CP( II ) ), CMIN )
           J = J+1
           B1( J ) = -2.0 +  H(Medium)**2 * omega2 / CP( II )**2
        END DO
     ELSE

        ! Case of an elastic medium

        IF ( SIGMA( Medium ) /= 0.0 ) THEN
           WRITE( PRTFIL, * ) 'Rough elastic interface not allowed'
           WRITE( PRTFIL, * ) 'PROGRAM ABORTING'
           STOP 'ERROR IN KRAKENC: Rough elastic interface not allowed'
        ENDIF
        Mater( Medium ) = 'ELASTIC'
        ELFLAG = .TRUE.

        DO II = 1, N1
           CMIN = MIN( DBLE( CS( II ) ), CMIN )
           J = J+1
           LAM = rho( J ) * ( CP( II ) ** 2 - 2.0 * CS( II )**2 )
           MU  = rho( J ) * CS( II ) ** 2

           B1( J ) = 1.0 / MU
           B2( J ) = 1.0 / ( LAM + 2.0 * MU )
           B3( J ) = 4.0 * MU * ( LAM + MU ) / ( LAM + 2.0 * MU )
           B4( J ) = LAM / ( LAM + 2.0 * MU )
           rho( J ) = omega2 * rho( J )

        END DO

     ENDIF
  END DO ! Next Medium:

  ! Bottom properties

  IF ( BOTOPT(1:1) == 'A' ) THEN
     IF ( CSB /= (0.0, 0.0) ) THEN ! Elastic bottom:
        Mater(NMedia+1) = 'ELASTIC'
        ELFLAG = .TRUE.
        CMIN = MIN( CMIN, DBLE(CSB) )
     ELSE                          ! Acoustic bottom:
        Mater(NMedia+1) = 'ACOUSTIC'
        CMIN = MIN( CMIN, DBLE(CPB) )
     ENDIF
  ENDIF

  ! Top properties

  IF ( TOPOPT(2:2) == 'A' ) THEN
     IF ( CST /= (0.0, 0.0) ) THEN ! Elastic top:
        ELFLAG = .TRUE.
        CMIN = MIN( CMIN, DBLE(CST) )
     ELSE                          ! Acoustic top:
        CMIN = MIN( CMIN, DBLE(CPT) )
     ENDIF
  ENDIF

  IF ( ELFLAG ) CMIN = 0.9*CMIN
  CLOW = MAX( CLOW, 0.99*CMIN )

  RETURN
END SUBROUTINE INIT
!**********************************************************************C
SUBROUTINE SOLVE( ERROR )

  ! Solves the eigenvalue problem at the current mesh
  ! and produces a new extrapolation

  USE krakelmod

  IMPLICIT REAL (KIND=8) (A-H, O-Z)
  REAL (KIND=8) :: T1, T2, F1, F2

  CALL CPU_TIME( Tstart )
  CALL INIT     ! set up the finite difference mesh                      
  CALL SOLVE2   ! solve for the eigenvalues

  Extrap( ISet, 1:M ) = EVMat( ISet, 1:M ) 
  IF ( ISet == 1 ) CALL VECTOR   ! If this is the first mesh, compute the eigenvectors

  ! Richardson extrapolation to improve the accuracy          

  ERROR = 10
  KEY = 2 * M / 3 + 1     ! index of element used to check convergence

  IF ( ISET > 1 ) THEN
     T1 = EXTRAP( 1, KEY )

     DO J = ISET, 2, -1
        DO MODE = 1, M
           F2 = EXTRAP( J,   MODE )
           F1 = EXTRAP( J-1, MODE )
           X2 = NV(ISET)**2
           X1 = NV(J-1)**2
           EXTRAP( J-1, MODE ) = F2 - (F1-F2)/(X2/X1-1.0)
        END DO
     END DO

     T2 = EXTRAP( 1, KEY )
     ERROR = ABS( T2 - T1 )
  ENDIF

  CALL CPU_TIME( Tend)   ! check elapsed time
  ET( ISet ) = Tend - Tstart
  WRITE( PRTFIL, '( 1X, I8, 6X, G15.3 )' ) NV( ISet ), ET( ISet )

  RETURN 
END SUBROUTINE SOLVE

!**********************************************************************C
SUBROUTINE SOLVE2

  ! Provides initial guess to root finder for each EVMAT(I)

  USE krakelmod

  IMPLICIT NONE
  INTEGER MaxIt, Med, NTot, II, J, IAllocStat, It
  REAL    (KIND=8) :: Tol
  REAL    (KIND=8) :: X, P(10)
  CHARACTER*80 ERRMSG

  X = omega2 / CLOW**2
  MaxIT = 500

  ! solve1 has already allocated space for the following unless the problem has shear
  IF ( .NOT. ALLOCATED( k ) ) THEN
     M = 3000   ! this sets the upper limit on how many modes can be calculated
     ALLOCATE( EVMAT( NSets, M ), Extrap( NSets, M ), k( M ), VG( M ), Stat = IAllocStat )
     IF ( IAllocStat /= 0 ) &
          CALL ERROUT( PRTFil, 'F', 'KRAKEN - SOLVE2', 'Insufficient memory (too many modes).' )
  END IF

  NTOT = SUM( N( NFact : NLact ) )

  DO MODE = 1, M

     ! For first or second meshes, use a high guess
     ! Otherwise use extrapolation to produce an initial guess

     X = 1.00001*X
     IF ( TOPOPT(4:4) == 'R' ) X = k(MODE)**2
     IF ( ISET >= 2 ) THEN

        DO II = 1, ISET-1
           P( II ) = EVMAT( II, MODE)
        END DO

        IF ( ISET >= 3 ) THEN
           DO II = 1, ISET-2
              DO J = 1, ISET - II - 1
                 P( J ) = ( ( HV( ISET )**2 - HV( J + II ) **2 ) * P( J ) - ( HV( ISET ) **2 &
                      -HV( J )**2 ) * P( J+1 ) ) / ( HV( J )**2 - HV( J + II ) **2 )
              END DO
           END DO
           X = P(1)
        ENDIF

     ENDIF

     ! Use the secant method to refine the eigenvalue
     Tol = ABS( X ) * 10.0 ** ( 4.0 - PRECISION( X ) )
     CALL ZSECX( X, Tol, IT, MaxIT, ERRMSG )

     IF ( ERRMSG /= ' ' ) THEN
        WRITE( PRTFIL, * ) ERRMSG
        WRITE( PRTFIL, * ) 'for ISET, MODE = ',ISET, MODE
     ENDIF

     EVMAT( ISET, MODE ) = X

     ! Toss out modes outside user specified spectral limits
     IF ( SQRT( omega2 ) / REAL( SQRT( X ) ) > CHIGH) THEN
        M = MODE - 1
        RETURN
     ENDIF

     IF ( MODE .GT. NTOT / 5 ) THEN
        WRITE( PRTFIL, * ) 'Caution: mesh is too coarse to compute all modes in spectrum'
        M = MODE
        RETURN
     ENDIF
  END DO   ! next mode

  ! Check whether we have hit the ceiling on max # modes

  IF ( M == MaxM ) THEN
     WRITE( PRTFIL, * ) 'Too many modes in phase speed interval--'
     WRITE( PRTFIL, * ) 'program will compute as many as it can'
  ENDIF

  RETURN
END SUBROUTINE SOLVE2
!**********************************************************************C
SUBROUTINE FUNCT( X, DELTA, IPOW )

  ! FUNCT(X) = 0 IS THE DISPERSION RELATION

  USE krakelmod
  IMPLICIT NONE
  INTEGER IPOW, NMat, lda, Ml, Mu, j, Info
  INTEGER IPVT( MaxN )
  REAL (KIND=8) :: A( 13, MaxN ), VDET( 2 )
  REAL (KIND=8) :: X, DELTA

  ! IF WE GET TO A K VALUE BELOW THE CTS SPECTRUM LIMIT
  ! THEN FORCE A ZERO BY RETURNING A NEW FUNCTION VALUE
  ! WHICH IS TEN ORDERS OF MAGNITUDE LESS THAN LAST VALUE

  IF ( REAL(X) <= omega2/CHIGH**2 ) THEN
     IPOW = IPOW-10
     RETURN
  ENDIF

  CALL SETA( A, NMat, X ) ! Set up the matrix

  lda = 13
  ML = 4
  MU = 4

  CALL DGBTRF( NMat, NMat, ML, MU, A, lda, IPVT, INFO )  ! Factor the matrix

  IF ( INFO /= 0 ) THEN
     WRITE(PRTFIL, *) 'INFO = ', INFO
     STOP
  ENDIF

  CALL DGBDI( A, LDA, NMat, ML, MU, IPVT, VDET )  ! Calculate the determinant

  DELTA = VDET( 1 )
  IPOW  = VDET( 2 )

  ! Deflate previous roots

  IF ( MODE > 1 ) THEN
     DELTA = DELTA / ( X-EVMAT( ISET, MODE-1 ) )
     IF ( MODE > 2 ) THEN
        DO J = 1, MODE-2
           DELTA = ( EVMAT( ISET, MODE - 1 ) - EVMAT( ISET, J ) ) * DELTA / &
                   ( X - EVMAT( ISET, J ) )
        END DO
     ENDIF
  ENDIF

  RETURN
END SUBROUTINE FUNCT
!**********************************************************************C
SUBROUTINE VECTOR

  ! Do inverse iteration to compute each of the eigenvectors
  ! and write these to the disk file

  USE krakelmod
  USE SdRdRMod

  IMPLICIT REAL (KIND=8) (A-H, O-Z)
  INTEGER IPVT( MaxN )
  REAL    (KIND=4) :: Z(MaxN)
  REAL    (KIND=8) :: A(13, MaxN), EV(MaxN), X
  COMPLEX (KIND=4) :: EVS(MaxN)
  CHARACTER BCTOP*1, BCBOT*1

  BCTop(1:1) = TopOpt(2:2) 
  BCBot(1:1) = BotOpt(1:1)

  ! Tabulate Z-coordinates & compute NMat, NTOT
  NMat = 0
  NTOT = 0
  J = 0

  DO Medium = 1, NMedia
     J = J+1
     Z( J ) = DEPTH( Medium )

     DO II = 1, N( Medium )
        J = J+1
        Z( J ) = Z( J-1 ) + H( Medium )
     END DO

     IF ( Mater( Medium ) == 'ACOUSTIC' ) THEN
        NMat = NMat + N( Medium )+1
     ELSE
        NMat = NMat + 4 * ( N( Medium ) + 1 )
     ENDIF

     NTOT = NTOT + N(Medium)+1
  END DO

  IF ( NMat .GT. 4095 ) THEN
     WRITE(PRTFIL, *) 'Too many depth points'
     WRITE(PRTFIL, *) 'PROGRAM ABORTING'
  ENDIF

  ! Open MODFIL and write header

  LRECL = MAX( 2 * NMat, 32, 3 * NMedia ) ! Logical record length in `longwords' (4 bytes)
  OPEN ( FILE = 'MODFIL', UNIT = MODFIL, ACCESS = 'DIRECT', RECL = 4 * LRECL, FORM = 'UNFORMATTED' )

  WRITE( MODFIL, REC = 1 ) LRECL, TITLE, REAL( FREQ ),  NMedia, NTOT, NMat
  WRITE( MODFIL, REC = 2 ) ( N( Medium ), Mater( Medium ), Medium = 1, NMedia )
  WRITE( MODFIL, REC = 3 ) BCTOP(1:1), CMPLX( CPT ), CMPLX( CST ), REAL( rhoT ), REAL( DEPTH( 1          ) ), &
       &   BCBOT(1:1), CMPLX( CPB ), CMPLX( CSB ), REAL( rhoB ), REAL( DEPTH( NMedia + 1 ) )
  WRITE( MODFIL, REC = 4 ) ( REAL( DEPTH( Medium ) ), REAL( rho( Loc( Medium ) + 1 ) ), Medium = NFACT, NLACT )
  WRITE( MODFIL, REC = 6 ) ( Z( IZ ), IZ = 1, NTOT )

  ! MAIN LOOP: For each eigenvalue call SINVIT to get eigenvector

  DO MODE = 1, M
     X = 1.00000001 * EVMAT(1, MODE)

     CALL SETA( A, NMat, X ) ! Set up the matrix

     LDA = 13
     ML  = 4
     MU  = 4
     CALL DGBTRF( NMat, NMat, ML, MU, A, lda, IPVT, INFO )  ! Factor the matrix

     IF ( INFO /= 0 ) THEN
        WRITE( PRTFIL, * ) 'INFO = ', INFO
        STOP
     ENDIF

     JOB = 0
     EV( 1: NMat ) = 1.0
     NRHS = 1      ! number of right hand sides
     ldb  = NMat   ! first dimension of b
     CALL DGBTRS( 'N', NMat, ML, MU, NRHS, A, LDA, IPVT, EV, ldb, JOB )  ! Solve
     CALL NORMIZ( EV, EVS, NMat, X ) ! Normalize

     WRITE( MODFIL, REC = 6 + MODE ) ( EVS( J ), J = 1, NMat )
  END DO   ! next mode

  RETURN
END SUBROUTINE VECTOR
!**********************************************************************C
SUBROUTINE SETA( A, NMat, X )

  ! ROUTINE FOR SETTING UP THE MATRIX A

  USE krakelmod
  IMPLICIT NONE
  INTEGER ii, j, l, Medium, NMat, IPow
  REAL    (KIND=8) :: A(13, MaxN), GamP, GamS, GamP2, GamS2, rmu, fac, hrho, Twobyh
  REAL    (KIND=8) :: X
  COMPLEX (KIND=8) :: F, G
  CHARACTER BCTYPE*1

  ! Compute size of matrix
  NMat = 0
  DO Medium = 1, NMedia
     IF ( Mater(Medium) == 'ACOUSTIC' ) THEN
        NMat = NMat + N( Medium ) + 1
     ELSE
        NMat = NMat + 4 * ( N( Medium ) + 1 )
     ENDIF
  END DO

  A = 0.0 ! Zero out the matrix

  ! MAIN LOOP ( OVER EACH Medium )

  J = 0
  DO Medium = 1, NMedia
     IF ( Medium /= 1 ) THEN

        ! Acousto-elastic interface

        IF ( Mater(Medium-1) == 'ACOUSTIC' .AND. Mater(Medium) == 'ELASTIC' ) THEN
           ! R2 Condition
           L = L + 1
           J = J + 1
           A( 10, J-1 ) = 1.0
           A(  9, J   ) = 0.5*( B1(L) - H(Medium-1)**2 * X )
           A(  7, J+2 ) = H(Medium-1)*omega2
           ! R3 Condition
           A( 10, J   ) = 0.0
           A(  7, J+3 ) = 1.0
           ! R4 Condition
           A( 11, J   ) = 1.0
           A(  7, J+4 ) = 1.0
        ENDIF

        ! Elasto-acoustic interface

        IF ( Mater(Medium-1) == 'ELASTIC' .AND. Mater(Medium) == 'ACOUSTIC' ) THEN
           L = Loc(Medium) + 1
           J = J + 5
           ! R3 Condition
           A(9, J-2) = 1.0
           ! R4 Condition
           A(9, J-1) = 1.0
           A(8, J  ) = 1.0
           ! R2 Condition
           A(12, J-3 ) = -H(Medium)*omega2
           A( 9, J   ) = 0.5*(B1(L)-H(Medium)**2*X)
           A( 8, J+1 ) = 1.0
        ENDIF

        ! Acoustic/acoustic interface

        IF ( Mater(Medium-1) == 'ACOUSTIC' .AND. Mater(Medium) == 'ACOUSTIC' ) THEN
           ! Continuity of normal displacement
           L = L + 1
           J = J + 1
           Hrho = H(Medium-1)*rho( Loc(Medium-1) + 1 )
           A( 10, J-1 ) = 1.0/Hrho
           A(  9, J   ) = 0.5*(B1(L)-H(Medium-1)**2*X)/Hrho
           L = L + 1
           Hrho = H(Medium)*rho( Loc(Medium) + 1 )
           A(  8, J+1 ) = 0.5*(B1(L)-H(Medium)**2*X)/Hrho
           A(  7, J+2 ) = 1.0/Hrho
           ! Continuity of pressure
           J = J + 1
           A( 10, J-1 ) = 1.0
           A(  9, J   ) = -1.0
        ENDIF

        ! Elastic/elastic interface

        IF ( Mater(Medium-1) == 'ELASTIC' .AND. Mater(Medium) == 'ELASTIC' ) THEN
           ! Continuity of everything
           J = J + 1
           A( 13, J-2 ) =  1.0
           A(  9, J+2 ) = -1.0
           J = J + 1
           A( 13, J-2 ) =  1.0
           A(  9, J+2 ) = -1.0
           J = J + 1
           A( 13, J-2 ) =  1.0
           A(  9, J+2 ) = -1.0
           J = J + 1
           A( 13, J-2 ) =  1.0
           A(  9, J+2 ) = -1.0
        ENDIF
     ENDIF

     IF ( Mater(Medium) == 'ACOUSTIC' ) THEN

        ! Acoustic section

        L = Loc(Medium) + 1
        IF ( Medium == 1 ) J = 1
        DO II = 1, N(Medium)-1
           J = J + 1
           L = L + 1
           A( 10, J-1 ) = 1.0
           A(  9, J   ) = B1(L) - H(Medium)*H(Medium)*X
           A(  8, J+1 ) = 1.0
        END DO
     ELSE

        ! Elastic section

        L = Loc(Medium)
        TWOBYH = 2.0/H(Medium)
        DO II = 1, N(Medium)
           L = L + 1
           J = J+1
           A( 11, J ) = TWOBYH
           A( 12, J ) = X*B4(L)
           A( 13, J ) = X*B3(L) - rho(L)
           J = J+1
           A( 10, J ) = -1.0
           A( 11, J ) = TWOBYH
           A( 13, J ) = -rho(L)
           J = J+1
           A(  9, J ) = B1(L)
           A( 11, J ) = TWOBYH
           A( 12, J ) = X
           J = J+1
           A(  9, J ) = B2(L)
           A( 10, J ) = -B4(L)
           A( 11, J ) = TWOBYH

           L = L + 1
           J = J + 1
           A( 7, J ) = -TWOBYH
           A( 8, J ) = X*B4(L)
           A( 9, J ) = X*B3(L) - rho(L)
           J = J+1
           A( 6, J ) = -1.0
           A( 7, J ) = -TWOBYH
           A( 9, J ) = -rho(L)
           J = J+1
           A( 5, J ) = B1(L)
           A( 7, J ) = -TWOBYH
           A( 8, J ) = X
           J = J+1
           A( 5, J ) = B2(L)
           A( 6, J ) = -B4(L)
           A( 7, J ) = -TWOBYH
           J = J-4
           L = L-1
        END DO
     ENDIF
  END DO   ! next medium

  ! Top BC

  IF ( Mater(1) == 'ELASTIC' ) THEN ! Elastic medium
     IF ( TOPOPT(2:2) == 'R' ) THEN ! Rigid top
        A( 9, 1 ) = 1.0
        A( 9, 2 ) = 1.0
     ENDIF
     IF ( TOPOPT(2:2) == 'V' ) THEN ! Free (vacuum) top
        A( 7, 3 ) = 1.0
        A( 7, 4 ) = 1.0
     ENDIF
  ELSE
     ! Acoustic medium
     ! Top impedance
     BCTYPE(1:1) = TOPOPT(2:2)
     CALL BCIMP( X, BCTYPE, 'TOP', CPT, CST, rhoT, F, G, IPOW )
     IF ( G == 0.0 ) THEN ! Free (vacuum) top
        A( 9, 1 ) = 1.0
        A( 8, 2) = 0.0
     ELSE
        Hrho = rho(1)*H(1)
        A( 9, 1 ) = 0.5*(B1(1)-H(1)**2*X)/Hrho + F/G
        A( 8, 2) = 1.0/Hrho
     ENDIF
  ENDIF

  ! Bottom BC

  IF ( Mater(NMedia) == 'ELASTIC' ) THEN ! Elastic medium
     IF ( BOTOPT(1:1) == 'R' ) THEN ! Rigid bottom
        A( 11, NMat-3 ) = 1.0
        A( 11, NMat-2 ) = 1.0
     ENDIF
     IF ( BOTOPT(1:1) == 'V' ) THEN ! Free (vacuum) bottom
        A( 9, NMat-1 ) = 1.0
        A( 9, NMat   ) = 1.0
     ENDIF

     IF ( BOTOPT(1:1) == 'A' ) THEN ! Elastic bottom
        GAMP2 = X - omega2/CPB**2
        GAMS2 = X - omega2/CSB**2
        GAMP = SQRT( GAMP2 )
        GAMS = SQRT( GAMS2 )
        RMU = rhoB*CSB**2
        FAC = -RMU / ( GAMS*GAMP - X )

        A( 11, NMat-3 ) = FAC*GAMP*(X-GAMS2)
        A( 10, NMat-2 ) = FAC*(2.0*GAMS*GAMP-(GAMS2+X))
        A( 12, NMat-3 ) = X*A(10, NMat-2)
        A( 11, NMat-2 ) = FAC*GAMS*(X-GAMS2)
        A(  9, NMat-1 ) = 1.0
        A(  9, NMat   ) = 1.0
     ENDIF
  ELSE
     ! Acoustic medium
     ! Bottom impedance
     BCTYPE(1:1) = BOTOPT(1:1)
     CALL BCIMP( X, BCTYPE, 'BOT', CPB, CSB, rhoB, F, G, IPOW )

     IF ( G == 0.0 ) THEN ! Free (vacuum) bottom
        A( 10, NMat-1 ) = 0.0
        A(  9, NMat   ) = 1.0
     ELSE
        L = L + 1
        Hrho = H(NMedia)*rho( Loc(NMedia) + 1 )
        A( 10, NMat-1) = 1.0/Hrho
        A(  9, NMat  ) = 0.5 * ( B1( L ) - H( NMedia )**2 * X ) / Hrho - F / G
     ENDIF
  ENDIF

  RETURN
END SUBROUTINE SETA
!**********************************************************************C
SUBROUTINE BCIMP( x, BCType, BotTop, CPHS, CSHS, rhoHS, F, G, IPow )                                                   

  ! Compute Boundary Condition IMPedance
  ! Same as KRAKENC except for ELASUP, ELASDN removed

  USE krakelmod
  use RefCoMod

  IMPLICIT REAL (KIND=8) (A-H, O-Z)
  INTEGER, INTENT( OUT ) :: IPow
  REAL    (KIND=8) :: rhoHS, x
  COMPLEX (KIND=8) :: cx, Kx, Kz, TWERSK, GamS2, GamP2, GamS, GamP, mu, RCmplx, CPHS, CSHS, CInside
  COMPLEX (KIND=8), INTENT( OUT ) :: F, G
  CHARACTER BCTYPE*1, BOTTOP*3

  cx = x   ! complex version of x for routines that expect a complex argument
  IPow = 0 

  ! Get rho, C just INSide the boundary               
  ! (There is at least one acoustic layer in the problem, except
  !  in the case where BOUNCE is used to get the refl. coef. for       
  !  a purely elastic stack of layers.                                 

  IF ( BotTop(1:3) == 'TOP' ) THEN 
     IF ( NFAct > 0 ) THEN
        Itop    = Loc( NFAct ) + N( NFAct ) + 1
        rhoINS  = rho( Itop )
        CInside = SQRT( Omega2 * H( NFAct ) **2 / ( 2.0 + B1( Itop ) ) )
     ENDIF
  ELSE 
     IF ( NLACT > 0 ) THEN 
        Ibot    = Loc( NLACT ) + N( NLACT ) + 1 
        rhoINS  = rho( Ibot )
        CInside = SQRT( Omega2 * H( NLACT ) **2 / ( 2.0 + B1( Ibot ) ) )
     ENDIF
  ENDIF

  ! Vacuum

  IF ( BCTYPE(1:1) == 'V' ) THEN
     F = 1.0
     ! G = -CI * PEKRT( omega2/CINS**2 - X ) * SIGMA(1)**2
     G = 0.0
  ENDIF

  ! Vacuum with Twersky scatter model

  IF ( BCTYPE(1:1) == 'S' .OR. BCTYPE(1:1) == 'H' .OR. &
       &     BCTYPE(1:1) == 'T' .OR. BCTYPE(1:1) == 'I' ) THEN

     omega = SQRT( omega2 )
     KX = SQRT( X )
     F = 1.0
     G = TWERSK( BCTYPE, omega, BUMDEN, XI, ETA, KX, rhoINS, CINS )
     G = G / ( i*omega*rhoINS )
  ENDIF

  ! Rigid

  IF ( BCTYPE(1:1) == 'R' ) THEN
     F = 0.0
     G = 1.0
  ENDIF

  ! Acousto-elastic half-space

  IF ( BCTYPE(1:1) == 'A' ) THEN
     GAMP = SQRT( X - omega2 / CPHS**2 )
     F = 1.0
     G = rhoB/GAMP
  ENDIF

  ! Tabulated reflection coefficient
  IF ( BCType(1:1) == 'F' ) THEN 
     ! Compute the grazing angle THETA
     Kx = SQRT( x ) 
     Kz = SQRT( Omega2 / CInside ** 2 - x ) 
     RadDeg = 360.0 / ( 2.0 * PI )
     ThetaInt = RadDeg * DATAN2( DBLE( Kz ), DBLE( Kx ) )

     ! Evaluate R( ThetaInt )
     IF ( BotTop(1:3) == 'TOP' ) THEN
        CALL REFCO( ThetaInt, RInt, PhiInt, ThetaTop, RTop, PhiTop, NTopPts, PRTFil )
     ELSE
        CALL REFCO( ThetaInt, RInt, PhiInt, ThetaBot, RBot, PhiBot, NBotPts, PRTFil )
     ENDIF

     ! Convert R(THETA) to (f,g) in Robin BC
     RCmplx = RInt * EXP( i * PhiInt )
     F      = 1.0
     G      = ( 1.0 + RCmplx ) / ( i * Kz * ( 1.0 - RCmplx ) )
  ENDIF

  IF ( BotTop(1:3) == 'TOP' ) G = -G    ! A top BC has the sign flipped relative to a bottom BC

  ! Precalculated reflection coef
  IF ( BCType(1:1) == 'P' ) THEN 
     CALL IRCINT( cx, F, G, IPow, xTab, FTab, GTab, ITab, NkTab ) 
  ENDIF

  RETURN
END SUBROUTINE BCIMP
!**********************************************************************C
SUBROUTINE NORMIZ( EV, EVS, NMat, X )

  ! Convert the EV to (U, W, TAUZX, TAUZZ) and normalize

  USE krakelmod
  IMPLICIT NONE
  INTEGER ii, Medium, NMat, j, kk
  REAL    (KIND=8) :: EV(MaxN), SqNorm2, rhoM
  COMPLEX (KIND=4) :: EVS(MaxN), CIK
  COMPLEX (KIND=8) :: R1, R2, R3, R4, X, SqNorm, Contrib, RN, ScaleFac, GAMP

  CIK = i * SQRT( X )

  ! Scale down the mode

  SqNorm2 = 0.0

  DO II = 1, NMat
     SqNorm2 = MAX( SqNorm2, ABS( EV( II ) ) )
  END DO

  EV( 1 : NMat ) = EV( 1 : NMat ) / SqNorm2

  ! Loop to compute norm of eigenvector

  SqNorm = 0.0
  J = 1
  KK = 1

  DO Medium = 1, NMedia
     rhoM = rho( Loc(Medium) + 1 )

     DO II = 1, N(Medium)+1
        IF ( Mater(Medium) == 'ELASTIC' ) THEN
           R1 = EV( KK )
           R2 = EV( KK+1 )
           R3 = EV( KK+2 )
           R4 = EV( KK+3 )

           Contrib = H( Medium ) * ( -X * B3(J) * R1**2 + B4(J) * R1 * R4 - R2 * R3)

           EVS( KK   ) = CIK * EV( KK )
           EVS( KK+1 ) =       EV( KK+1 )
           EVS( KK+2 ) = CIK * EV( KK+2 )
           EVS( KK+3 ) =       EV( KK+3 )

           KK = KK+4
        ELSE
           Contrib = -H(Medium)*EV(KK)**2/(rhoM*omega2)
           EVS( KK ) = -EV( KK )
           KK = KK+1
        ENDIF

        IF ( II == 1 .OR. II == N(Medium)+1 ) Contrib = 0.5*Contrib
        SqNorm = SqNorm + Contrib
        J = J+1
     END DO
  END DO

  ! Bottom half-space contribution

  IF ( BOTOPT(1:1) == 'A' ) THEN
     IF ( Mater(NMedia+1) == 'ELASTIC' ) THEN
        WRITE( PRTFIL, * ) 'Elastic halfspace normalization not implemented'
     ELSE
        GAMP = SQRT( X - omega2 / CPB**2 )
        Contrib = -EV( NMat )**2 / ( 2.0 * GAMP * rhoB * omega2 )
     ENDIF
  ENDIF

  SqNorm = SqNorm + Contrib

  ! Scale the mode

  RN = -omega2 * SqNorm
  ScaleFac = 1.0 / SQRT( RN )
  EVS( 1 : NMat ) = ScaleFac * EVS( 1 : NMat )

  RETURN
END SUBROUTINE NORMIZ
