function P = evalcmLoadAll( phiS, rd, Nrd, r, Nr, M, Opt )

% conversion from Fortran routine evalcm.f90
% mbp 4/2009


% Computes pressure field using coupled mode theory
% Normalized to pressure of point source at 1 meter

% Opt = X     Cartesian   (X, z) coordinates
% Opt = R     Cylindrical (R, z) coordinates

% Note number of propagating modes is reset after first segment.
% Thus M restricts the number of modes in the source field but
% thereafter energy is allowed to couple into higher-order modes.

% Should half-space contribution include imaginary part?
% Then we need to handle the growing part of leaky waves.

%   LOGICAL   first

global xBot

persistent first
%   DATA first /.TRUE./

Dwedge = linspace( 3000, 500, 101 );

P = zeros( Nrd, Nr );   % pre-allocated pressure field

% load all the mode information
MaxM = 700
phiR = zeros( Nrd, MaxM, length( Dwedge ) );

for ii = 1 : length( Dwedge )
   filename = sprintf( 'MODFIL%04i', ii );
   fprintf( 'Reading mode file # %4i \r', ii );
   
   Modes( ii ) = read_modes( filename );
   
   % Subtabulate modes at receiver depths
   phiR( :, 1:size( Modes( ii ).phi, 2 ), ii ) = interp1q( Modes( ii ).z, Modes( ii ).phi, rd' );   % subsample phi onto the receiver grid

   %    if ( MR == 1 )   % interp1 converted phiR to a row vector if MR == 1
   %       phiR = phiR.';
   %    end
   %
   
   if ( Modes( ii ).Top.bc == 'A' )
      iTop = find ( rd < Modes( ii ).Top.depth ); % Rcvr in upper halfspace
      phiR( iTop, 1 : Modes( ii ).M ) = exp( -( Modes( ii ).Top.depth   - rd( iTop  ) ).' * Modes( ii ).Top.gamma.' ) * diag( Modes( ii ).Top.phi );
   end
   
   if ( Modes( ii ).Bot.bc == 'A' )
      iBot = find ( rd > Modes( ii ).Bot.depth ); % Rcvr in lower halfspace
      phiR( iBot, 1 : Modes( ii ).M ) = exp( -( rd( iBot )  - Modes( ii ).Bot.depth   ).' * Modes( ii ).Bot.gamma.' ) * diag( Modes( ii ).Bot.phi );
   end
   
end

k   = Modes( 1 ).k;
M1  = length( k );
M   = min( M, M1 );
k   = k( 1 : M );   % reduce size of vector based on Mlimit

% Initialize the amplitudes A
if ( Opt(1:1) == 'X' )   % Cartesian coordinates
   A =      sqrt( 2.0 * pi ) * exp( 1i * pi / 4.0 ) * phiS( 1:M ) ./       k( 1:M );
else                     % Cylindrical coordinates
   A = 1i * sqrt( 2.0 * pi ) * exp( 1i * pi / 4.0 ) * phiS( 1:M ) ./ sqrt( k( 1:M ) );
end


%A = [ zeros( length( A ) - 1, 1 ); 1 ];

% March forward in range
for ir = 1 : Nr
   % calculate the bottom depth at this range
   D = interp1q( xBot( 1, : )', xBot( 2, : )', r( ir ) );
   iDnew = find ( D > Dwedge );
   if ( isempty( iDnew ) )   % didn't find any wedge depths above this?
      iDnew( 1 ) = length( Dwedge );
   end
   
   if ( ir == 1 )   % first receiver range is a special case
      iD = iDnew( 1 );
      phiTemp = phiR( :, 1:M, iD );         % get modes at receiver depths
      A = A .* exp( -1i * k * r( ir ) );                       % advance the phase of the coefficients
      
   else
      if ( iDnew( 1 ) ~= iD )   % Crossing into new profile?
         % calculate range of the interface
         alpha = ( D                        - Modes( iD ).Bot.depth ) / ...
            ( Modes( iDnew( 1 ) ).Bot.depth - Modes( iD ).Bot.depth ); % proportional distance
         rint = alpha * r( ir - 1 ) + ( 1 - alpha ) * r( ir ); % range of the interface
         
         A = A .* exp( -1i * k * ( rint - r( ir - 1 ) ) );     % Advance to the interface
         
         [ A, k, M ] = NewProfile( Modes( iD ), Modes( iDnew( 1 ) ), M, A );
         %disp( 'New profile read', r(ir ), iProf, ' #modes = ', M )
         
         A = A .* exp( -1i * k * ( r( ir ) - rint ) );         % Advance the remaining distance
         iD = iDnew( 1 );
         
         phiTemp = phiR( :, 1:M, iD );                         % get new modes at receiver depths
      else
         A = A .* exp( -1i * k * ( r( ir ) - r( ir - 1 ) ) );  % advance the phase of the coefficients
      end
      
   end

   % Add up modal contributions
   if ( M == 0 )    % modes cut off?
      P( 1, 1, :, ir ) = 0;
   else
      if ( Opt(1:1) == 'R' && r( ir ) ~= 0.0 )
         P( :, ir ) =  phiTemp * A / realsqrt( r( ir ) );
      else
         P( :, ir ) =  phiTemp * A( 1:M );
      end
   end
   

end    % next range step

P = reshape( P, 1, 1, Nrd, Nr );

%**********************************************************************C
function [ A, kR, MR ] = NewProfile( ModesL, ModesR, M, A )

% For a given profil number:
%     read in modes for current segment
%     project the pressure field onto the new modes
%     extract values of the modes at rcvr depths

% Compute pressure along the left of the interface and read new mode set

if ( M == 0 ) % quick return if no propagating modes
   A     = [];
   kR    = [];
   MR    = 0;
   return
end

[ P, z, kR, NR, NTot, ML, MR, gamTL, gamBL, depthTL, depthBL, phiTL, phiBL ] = ...
   pleft( ModesL, ModesR, A, M );

% Calculate new amplitudes by projecting pressure onto the new modes
if ( MR == 0 ) % quick return if no propagating modes
   return
end

A = zeros( MR, 1 );   % shrink size of A as necessary

% calculate modal values for top and bottom halfspaces

if ( ModesR.Top.bc == 'A' )
   phiTR  = ModesR.phi( 1, 1:MR );
%    gamma2 = ModesR.k( 1: MR ) .^ 2 - ModesR.Top.k2;
%    gamTR  = PekerisRoot( gamma2 );
   gamTR = ModesR.Top.gamma;
end

if ( ModesR.Bot.bc == 'A' )
   %phiBR  = ModesR.phi( NR, 1:MR );
   phiBR  = ModesR.Bot.phi;
   gamBR  = ModesR.Bot.gamma;
end

% get indices of depths in halfspaces
if ( ModesR.Top.bc == 'A' )
   iTop = find ( z < ModesR.Top.depth );
end

if ( ModesR.Bot.bc == 'A' )
   iBot = find ( z > ModesR.Bot.depth );
end

if ( ModesR.Top.bc == 'A' )  % contribution from upper halfspace
   sum.Top.Tails   = tail( z( 1    ), phiTL, gamTL, depthTL, ...
      phiTR( mode ) / ModesR.Top.rho, gamTR( mode ), ModesR.Top.depth );
end

if ( ModesR.Bot.bc == 'A' )  % contribution from lower halfspace
   sum.Bot.Tails   = tail( z( NTot ), phiBL, gamBL, depthBL, ...
      phiBR / ModesR.Bot.rho, gamBR, ModesR.Bot.depth );
end


for imode = 1 : MR
   % tabulate the new mode on the grid from the previous segment
   phiTmp = ModesR.phi( :, imode );
   
   if ( ModesR.Top.bc == 'A' )
      phiTmp( iTop ) = phiTR( imode ) * exp( -gamTR( imode ) * ( ModesR.Top.depth - z( iTop ) ) );
   end
   
   if ( ModesR.Bot.bc == 'A' )
      phiTmp( iBot ) = phiBR( imode ) * exp( -gamBR( imode ) * ( z( iBot ) - ModesR.Bot.depth ) );
   end
   
   % Compute new amplitudes:
   %    A = Integral[ P( z ) * phi( z ) dz ]
   %    Integral is done using trapezoidal rule
   
   sum1 = P.' * phiTmp + sum.Bot.Tails( imode );
   %disp( [ mode, abs( sum1 ) ] )
   
   %disp( [ mode, abs( sum1 ) ] )
   A( imode ) = sum1;
end    % next mode



%abs( A( end ) )

fprintf( 'depth-averaged power: %10.2e \n', norm( A( 1:MR ) )^2 )

%**********************************************************************C

function [ P, z, kR, NR, NTot, ML, M, gamTL, gamBL, depthTL, depthBL, phiTL, phiBL ] = ...
   pleft( ModesL, ModesR, A, ML )

% Computes the pressure field along the interface using the depth sampling of the
% next segment.
% Also returns information needed for the tails in the halfspaces

% Get modal info at end of last segment (stored in structure Modes)

zL      = ModesL.z;
depthBL = ModesL.Bot.depth;
depthTL = ModesL.Top.depth;
ML      = min( length( ModesL.k ), ML );   % we only know M amplitudes
A       = A( 1 : ML );
Modes.k = ModesL.k( 1 : ML );
NL      = length( ModesL.z );

% Halfspace information
phiTL = A.' .* ModesL.phi( 1,  1 : ML );
phiBL = A.' .* ModesL.phi( NL, 1 : ML );

gamTL = zeros( 1, ML );
gamBL = zeros( 1, ML );

% we take the real part, gamma2, since leaky modes grow in depth
if ( ModesL.Top.bc == 'A' )
   gamTL  = ModesL.Top.gamma;
end

if ( ModesL.Bot.bc == 'A' )
   gamBL  = ModesL.Bot.gamma;
end

phiTemp = ModesL.phi( :, 1 : ML );
PL = phiTemp * A;   % Compute pressure at the interface

z  = ModesR.z;

NR = length( z );
M  = ModesR.M;

if ( M == 0 )   % jump out if modefile was empty
   kR   = [];
   NR   = [];
   NTot = [];
   P    = zeros( size( z ) );
   return
end


% Upslope? Extend the z vector with data from zL
NTot = NR;

ii   = find( zL > ModesR.z( NTot ) );
newz = length( ii );

z( NTot + 1 : NTot + newz ) = zL( ii );
NTot = NTot + newz;

kR = ModesR.k( 1 : M );
if ( ModesR.z( 1 ) ~= ModesR.Top.depth || ModesR.z( end ) ~= ModesR.Bot.depth )
   disp( 'Fatal Error: modes must be tabulated throughout the ocean and sediment to compute the coupling coefficients' )
   stop
end

% Tabulate medium density on new grid

rhomed = zeros( size( z ) );

for med = 1 : ModesR.Nmedia - 1
   ii =  z >= ModesR.depth( med ) & z <= ModesR.depth( med + 1 ) ;
   rhomed( ii ) = ModesR.rho( med );
end

% special case for last medium
ii =  z >= ModesR.depth( ModesR.Nmedia ) ;
rhomed( ii ) = ModesR.rho( ModesR.Nmedia );

% special case for upper halfspace
ii =  z < ModesR.Top.depth ;
rhomed( ii ) = ModesR.Top.rho;

% special case for lower halfspace
ii =  z > ModesR.Bot.depth ;
rhomed( ii ) = ModesR.Bot.rho;

% interpolate pressure onto the new grid
P = interp1q( zL, PL, z );

% special case for upper halfspace
if ( ModesR.Top.bc == 'A' )
   ii = find( z < depthTL );
   if ( ~isempty( ii ) )
      P( ii ) = phiTL(1:ML) * exp( -gamTL(1:ML) * ( depthTL - z( ii )' ) );
   end
end

% special case for lower halfspace
if ( ModesR.Bot.bc == 'A' )
   ii = find( z > depthBL );
   if ( ~isempty( ii ) )
      P( ii ) = phiBL(1:ML) * exp( -gamBL(1:ML) * ( z( ii )' - depthBL ) );
   end
end

% compute mesh width, h (it's actually h/rho)

% check this is correct for an irregular mesh

h = zeros( size( z ) );

h( 1           ) = 0.5 * ( z( 2       ) - z( 1           ) )  / rhomed( 1           ); % forward  difference
h( 2 : end - 1 ) = 0.5 * ( z( 3 : end ) - z( 1 : end - 2 ) ) ./ rhomed( 2 : end - 1 ); % centered difference
h( NTot        ) = 0.5 * ( z( NTot    ) - z( NTot - 1    ) )  / rhomed( NTot        ); % backward difference

% Point just above or below the interface
for med = 2 : ModesR.Nmedia
   % find points where the centered difference straddles an interface
   ii = find( z( 1: end - 2 ) < ModesR.depth( med ) & z( 3 : end ) > ModesR.depth( med ) ) + 1;
   DBelow  = ModesR.depth( med );
   
   h( ii ) = 0.5 * ( z( ii + 1 ) / rhomed( ii + 1 ) - z( ii - 1 ) / rhomed( ii - 1 ) ...
                        - DBelow / rhomed( ii + 1 ) +      DBelow / rhomed( ii - 1 ) );
end

P = h .* P;   % weight P by that h for later use in Trapezoidal rule


%**********************************************************************C
function [ tail ] = tail( D, phiL, gamL, DL, phiR, gamR, DR )

% Computes the contributions from the tails in the halfspaces
% gamL and phiL are vectors

tail = zeros( length( phiR ), 1 );

FL = phiL .* exp( -gamL * ( D - DL ) ).';
FR = phiR .* exp( -gamR * ( D - DR ) ).';

gammaInv = 1 ./ ( gamL * ones( 1, length( gamR ) )  + ones( length( gamL ), 1 ) * gamR.' );
tail = FR .* ( FL * gammaInv );


