function [ Modes ] = read_modes_bin( filename, modes )

% Read the modes produced by KRAKEN
% usage:
%    [ pltitl, freq, ck, z, phi ] = read_modes_bin( filename, modes )
% filename is without the extension, which is assumed to be '.moA'
% modes is an optional vector of mode indices

% derived from readKRAKEN.m    Feb 12, 1996 Aaron Thode
%
% Modes.M          number of modes
% Modes.k          wavenumbers
% Modes.z          sample depths for modes
% Modes.phi        modes
% Modes.Top.bc
% Modes.Top.cp
% Modes.Top.cs
% Modes.Top.rho
% Modes.Top.depth
% Modes.Bot.bc
% Modes.Bot.cp
% Modes.Bot.cs
% Modes.Bot.rho
% Modes.Bot.depth
% Modes.Nmedia     Number of media
% Modes.depth      depths of interfaces
% Modes.rho        densities in each medium

fid   = fopen( filename, 'r' );
lrecl = fread( fid, 1, 'long' );

%Try changing this line for the sun, if doesn't work
lrecl = 4 * lrecl;

rec = 0;  %Record length is one less than in the krakm.f file
fseek( fid, rec * lrecl, -1 );
fseek( fid, 4, 0 );
Modes.title  = char( fread( fid, 80, 'uchar' ) )';
Modes.freq   = fread( fid, 1, 'float' );
Modes.Nmedia = fread( fid, 1, 'long' );
Ntot         = fread( fid, 1, 'long' );

if Ntot < 0, return; end

% N and Mater
rec = 1;
fseek( fid, rec * lrecl, -1 );
N     = fread( fid, Modes.Nmedia, 'long' );
Mater = char( fread( fid, [ 8, Modes.Nmedia ], 'uchar' ) );

% Top
rec = 2;
fseek( fid, rec * lrecl, -1 );
Modes.Top.bc    = char( fread( fid, 1, 'char' ) );
cp              = fread( fid, [ 2, 1 ], 'float' );
Modes.Top.cp    = complex( cp( 1 ), cp( 2 ) );
cs              = fread( fid, [ 2, 1 ], 'float' );
Modes.Top.cs    = complex( cs( 1 ), cs( 2 ) );
Modes.Top.rho   = fread( fid, 1, 'float' );
Modes.Top.depth = fread( fid, 1, 'float' );

% Bottom
Modes.Bot.bc    = char( fread( fid, 1, 'char' )' );
cp              = fread( fid, [ 2, 1], 'float' );
Modes.Bot.cp    = complex( cp( 1 ), cp( 2 ) );
cs              = fread( fid, [ 2, 1], 'float' );
Modes.Bot.cs    = complex( cs( 1 ), cs( 2 ) );
Modes.Bot.rho   = fread( fid, 1, 'float' );
Modes.Bot.depth = fread( fid, 1, 'float' );

% depth and rho
rec = 3;
fseek( fid, rec * lrecl, -1 );
bulk        = fread( fid, [ 2, Modes.Nmedia ], 'float' );
Modes.depth = bulk( 1, : );
Modes.rho   = bulk( 2, : );

% m
rec = 4;
fseek( fid, rec * lrecl, -1 );
Modes.M = fread( fid, 1, 'long' );
Lrecl   = fread( fid, 1, 'long' );

% z
rec = 5;
fseek( fid, rec * lrecl, -1 );
Modes.z = fread( fid, Ntot, 'float' );

% read in the modes

if nargin == 1
   modes = 1:Modes.M;    % read all modes if the user didn't specify
end

% don't try to read modes that don't exist
ii    =  modes <= Modes.M ;
modes = modes( ii );

%Modes.phi = zeros( Ntot, length( modes ) );   %number of modes

% for ii = 1: length( modes )
%     rec = 5 + modes( ii );
%     fseek( fid, rec * lrecl, -1 );
%     phitmp = fread( fid, [ 2, Ntot ], 'float' )'; %Data is read columwise
%     Modes.phi( :, ii ) = phitmp( :, 1 ) + 1i * phitmp( :, 2 );
% end

rec = 6;
fseek( fid, rec * lrecl, -1 );

% if there are modes, read them
if ( Modes.M == 0 )
   Modes.phi = [];
   Modes.k   = [];
else
   phitmp    = fread( fid, [ 2, Ntot * Modes.M ], 'float' ); %Data is read columwise
   phitmp    = phitmp( 1, : ) + 1i * phitmp( 2, : );
   Modes.phi = reshape( phitmp, Ntot, Modes.M );
   
   % read in the wavenumbers
   
   % Ifirst = 1;
   % cktot = [];
   
   %for I = 1 : ( 1 + ( 2 * m - 1 ) / Lrecl ),
   %    rec = 5 + m + I;
   %    fseek( fid, rec * lrecl, -1 );
   %    Ilast = min( [ m Ifirst + Lrecl / 2 - 1 ] );
   %    ck = fread( fid, [ 2, Ilast - Ifirst + 1 ], 'float' )';
   %    cktot = [ cktot; ck ];
   %    Ifirst = Ilast + 1;
   %end
   %ck = cktot( modes, 1 ) + i * cktot( modes, 2 );
   
   rec = 6 + Modes.M;
   fseek( fid, rec * lrecl, -1 );
   k       = fread( fid, [ 2, Modes.M ], 'float' );
   k       = k( 1, : ) + 1i * k( 2, : );
   Modes.k = k(:);    % return a column vector
end

fclose( fid );
