function [vr,rhor,rhob,rhow,delta,Refl,Refl_dB,theta,ssbot,Cw] = Mz2GAP(Mz,Cw);
%MEAN_GR2_GEO   Convert grain size to geoacoustic properties
%               (sound speed, density and attenuation) and
%               generate the bottom reflection loss and write to
%               file. The equations for this are taken from the
%               Applied Physics Laboratory University of Washington
%               Technical Report 9407, Seattle, Washington.
%
%INPUTS        Mz   ==> Mean grain size with Mz = -log2(d/d0) with
%                       d as the mean grain diameter and d0 the
%                       reference diameter of 1 mm.
%
%              Cw ==> Reference sound speed (at water-sediment
%                       interface (m/s).
%
%OUTPUTS       vr   ==> Ratio of sound speed in bottom to Cw.
%              rhor ==> Density of bottom (assuming rho=1 for water).
%              delta==> Loss parameter is the ratio of the imaginary
%                       to real part of acoustic wavenumber (k).
%              Refl ==> Complex value reflection loss (function of angle)
%              theta==> Angle corresponding to Refl
%
%EXAMPLE
%
%   [vr,rhor,delta,Refl,theta] = mean_gr2_geo(3,1500);
%   plot(plot(theta,-20*log10(abs(Refl)))

%Define Cw to 1500 if not specified
if(nargin == 1)
    Cw = 1500;
    disp('Using 1500 m/s reference')
end

%m/ms for consistency with Hamilton (see APL Report)
c1 = Cw./1000;

%Equations from chapter 4 in APL Report
for n = 1:length(Mz)
    
    if(Mz(n) >= -1 & Mz(n) < 1)
        vr(n) = 0.002709.*Mz(n).^2 - 0.056452.*Mz(n) + 1.2778;
        rhor(n) = 0.007797.*Mz(n).^2 - 0.17057.*Mz(n) + 2.3139;
    elseif(Mz(n) >= 1 & Mz(n) < 5.3)
        vr(n) = -0.0014881.*Mz(n).^3 + 0.0213937.*Mz(n).^2 - 0.1382798.*Mz(n) + 1.3425;
        rhor(n) = -0.0165406.*Mz(n).^3 + 0.2290201.*Mz(n).^2 - 1.1069031.*Mz(n) + 3.0455;
    else
        vr(n) = -0.0024324.*Mz(n) + 1.0019;
        rhor(n) = -0.0012973.*Mz(n) + 1.1565;
    end;
    
    if(Mz(n) >= -1 & Mz(n) < 0)
        alf2_f(n) = 0.4556;
    elseif(Mz(n) >= 0 & Mz(n) < 2.6)
        alf2_f(n) = 0.4556 + 0.0245.*Mz(n);
    elseif(Mz(n) >= 2.6 & Mz(n) < 4.5)
        alf2_f(n) = 0.1978 + 0.1245.*Mz(n);
    elseif(Mz(n) >= 4.5 & Mz(n) < 6.0)
        alf2_f(n) = 8.0399 - 2.5228.*Mz(n) + 0.20098.*Mz(n).^2;
    elseif(Mz(n) >= 6.0 & Mz(n) < 9.5)
        alf2_f(n) = 0.9431 - 0.2041.*Mz(n) + 0.0117.*Mz(n).^2;
    else
        alf2_f(n) =  0.0601;
    end;
    delta(n) = alf2_f(n).*vr(n).*c1.*log(10)./(40*pi);
end;

theta = [0:1:90];
disp(['Calculating reflection loss for Mz = ',num2str(Mz(n))])
[Refl,Refl_dB] = mk_refl_loss(theta,vr(n),rhor(n),delta(n));
% % %Noticed some difference in the reflection loss curves
% % %may be due to the approximation of adding small imag part to c rather than
% % %k
rhow = 1.028;
rhob = rhor.*rhow;

ssbot = vr*Cw;
delta_wavl = delta.*54.58;
BHparms.bottom.props = [1 0 ssbot 0 rhor delta_wavl  0];
% BHparms.ssps.wat = [0 Cw];          %specify water column sound speed (at interface)
% BHparms.freqs = 200; %frequencies
theta = [0:1:90];             %angles in the reflection loss calculation
%[RLGT] = ReflLoss_wlayers(BHparms,theta);     %calculate refl. loss

%function for reflection loss calculated using APL Report
function [Refl,Refl_dB] = mk_refl_loss(theta,vr,rhor,delta);

kappa = (1./vr).*(1+i.*delta);
thetacon = theta.*pi/180;

for n = 1:length(theta)
    P(n) = sqrt(kappa.^2 - cos(thetacon(n)).^2);
    y = rhor.*sin(thetacon(n))./P(n);
    R(n) = (y - 1)./(y + 1);
    Refl_dB(n) = -20*log10(abs(R(n)));
    Refl(n) = R(n);
end;
%Make a BRCFIL for Bellhop
disp('Making BRCFIL')
fid = fopen('BRCFIL','w');
fprintf(fid,'%i\n', length(theta));
for n = 1:length(theta)
    fprintf(fid,'%4.4f %4.4f %4.4f\n', [theta(n) abs(Refl(n)) angle(Refl(n))*180/pi]);
end;
fclose(fid);
